//
// BMP Viewing Module
//
// 02/09/2023: Switch for bitmap depth
// 02/09/2023: Should fail more gracefully if not enough memory
// 02/09/2023: Tidy declarations / initialisations
//

#include <cybiko.h>
#include "bmp.h"

struct bitmap_info bminfo;
struct bitmap_header bmheader;

struct FileInput bmp_file_input;

struct
{
char r;
char g;
char b;
}col;

/***/

long swp(long);
int bswp(int);
char gbyte24(char r,char g,char b);
char grey24(int r,int g,int b);

void zoom(char*,char*,int,int,int,int);
void ordered(char *,char *,int,int);

void gauge(struct module_t *,long,long);
int waitinput(struct module_t *);

/***/


void plot(char *scr,char *dis,long w,long h,long sx,long sy){
	//sx and sy are scroll offsets;
	long pos;
	int j;
	int ah=100;	
	int aw=(w & 0xfffc)>>2;

	// Scroll width
	int sw=aw;
	if (sw>40){sw=40;}
	if (h<100){ah=h;};

	pos=(aw*sy)+sx;

	for (j=0;j<ah;j++){
		memcpy(dis,scr+pos,sw);
		pos+=aw;
		dis+=40;
	}
}


/* ************************ */

//
// Bitmap rows need to be padded so the are divisible by 4
// This function returns the padding value.

int padding(long x){
	while(x>0)
	{x-=4;}
	if (x<0){x=-x;}
	
	TRACE("Padding:%ld",x);

	return x;
}


//
// 1 Bit bitmap decoder. This is a seperate function as it is also used by
// the html viewer.

long bmp1(char *name,char *dest,int y,char htm){
	struct bitmap_info bminfo;
	struct bitmap_header bmheader;

	long bx,by=0,bw;
	int bd;
	
	FileInput_ctor(&bmp_file_input);

	if( FileInput_open( &bmp_file_input,name)){
		//
		Input_read( &bmp_file_input, &bmheader,14);
		Input_read( &bmp_file_input, &bminfo,40);

		bx=swp(bminfo.width);
		by=swp(bminfo.height);
		bd=bswp(bminfo.depth);

		TRACE("bmp: %ld x %ld x %ld",bx,by,bd);

		if (bd==1){
			char jk[8];
			int i,j,dif=0;
			
			// find the number of bytes used (8 pixels = 1 byte)
			int range=bx;
			while(range>0){
				range-=8;
				dif++;
			}

			TRACE("Range:%d",range);
			
			// get the byte padding
			dif=padding(dif);

			//
			if (htm==1){
				bw=40;
			}else{
				bw=(bx>>2);
			}

			// skip the colour part
			FileInput_seek( &bmp_file_input, 8, SEEK_CUR );

			for(j=0;j<by;j++){
				long pos=(y+by-j-1)*bw;

				 /* bx divided by 8*/
				for(i=0;i<bx;i+=8){
					char bit=FileInput_read_byte( &bmp_file_input);        	         
					
					char byt=0;
					if ((bit & 0x80)!=0x80){byt=192;}
					if ((bit & 64)!=64){byt+=48;}
					if ((bit & 32)!=32){byt+=12;}
					if ((bit & 16)!=16){byt+=3;}		
					dest[pos]=byt;
					pos++;
					
					byt=0;
					if ((bit & 8)!=8){byt=192;}
					if ((bit & 4)!=4){byt+=48;}
					if ((bit & 2)!=2){byt+=12;}
					if ((bit & 1)!=1){byt+=3;}		
					dest[pos]=byt;
					pos++;
				}

				Input_read(&bmp_file_input,&jk,dif);	         
			}
		}
	//
	}

	FileInput_dtor( &bmp_file_input, LEAVE_MEMORY);
	return by;
}


/* ************************ */

void bmp(char *myfile,struct module_t *ptr_main_module){
	long bx,by,bd;
	char *scr,*dis,*grey,*row;
	char abort=0;

	dis=DisplayGraphics_get_page_ptr(ptr_main_module->m_gfx, 0); 

	// Use screen memory to hold grey row (temp buffer)
	grey=dis;
	row=(dis+1000);

	/* ------------------------- */

	FileInput_ctor(&bmp_file_input);

	if(!FileInput_open( &bmp_file_input,myfile  ))
	{
		TRACE("Could not open file.\n");
	}else{
		memset( dis, 0, 4000 );

		Input_read( &bmp_file_input, &bmheader,14);
		Input_read( &bmp_file_input, &bminfo,40);

		bx=swp(bminfo.width);
		by=swp(bminfo.height);
		bd=bswp(bminfo.depth);
		TRACE("%ld x %ld x %ld",bx,by,bd);
		
		/** BITMAP ACTUAL SIZE (+PADDING) **/

		scr=(char *)calloc(1,(bx*by)>>2 );
		if (scr!=NULL){
			int i,j,dif=0;
			char jk[8];
			
			switch(bd){
				case 1:
					// 1bit Mono 
					bmp1(myfile,scr,0,0);
				break;
				
				case 4:
					// get the byte padding
					dif=padding( (bx>>1) );

					// Read the palette
					for(i=0;i<16;i++){	         
						Input_read(&bmp_file_input,&col,3);
						row[i]=grey24(col.r,col.g,col.b);	
						Input_read(&bmp_file_input,&jk,1); // padding/alpha        
					}

					for(j=by;j>0;j--){
						for(i=0;i<bx;i=i+2){
							char byt=FileInput_read_byte( &bmp_file_input);        	         
							char c1=(byt & 0x000f);
							char c2=(byt & 0x00f0)>>4;

							grey[i]=row[c2];
							grey[i+1]=row[c1];
						}

						ordered(scr,grey,bx,j);
						Input_read(&bmp_file_input,&jk,dif);	         
						gauge(ptr_main_module,(by-j),by);	
					}
				break;
				
				case 24:
					// get the byte padding
					dif=padding( (bx*3) );

					for(j=by;j>0;j--){
						Input_read(&bmp_file_input,row,(bx*3));

						for(i=0;i<bx;i++){
							grey[i]=grey24(row[(i*3)],row[(i*3)+1],row[(i*3)+2]);		
						}

						ordered(scr,grey,bx,j);
						gauge(ptr_main_module,(by-j),by);		
					}
				break;
				
			}
		
		}else{
			TRACE("Not Enough Memory (Required: %d bytes)..",((bx*by)>>2) );
			abort=1;
		}

		FileInput_dtor( &bmp_file_input, LEAVE_MEMORY);

		if (abort==0){
			long sx=0;
			long sy=0;
			int zm=0;
			// -------------

			int wmax,hmax;
			if (by>100){hmax=100;}else{hmax=by;}
			if (bx>160){wmax=160;}else{wmax=bx;}

			TRACE("Displaying.");
			do{
				memset(dis,0,4000);

				if (sx<0){sx=0;}
				if (sy<0){sy=0;}
					
				if (zm==1){
					if (sx>((bx>>2)-20)){sx=(bx>>2)-20;}
					if ((sy+50)>by){sy=by-50;}

					// Show zoomed
					zoom(scr,dis,bx,by,sx,sy);
				}else{
					if ((sx<<2)+wmax>bx){sx=(bx>>2)-(wmax/4);}
					if ((sy+hmax)>by){sy=by-hmax;}

					// Show normal
					plot(scr,dis,bx,by,sx,sy);
				}
			 
				DisplayGraphics_show( ptr_main_module->m_gfx );

				//--
				switch(waitinput(ptr_main_module)){
					case KEY_ESC:
						abort=1;
					break;

					case KEY_UP:
					case KEY_Q:
						sy-=4;
					break; 

					case KEY_DOWN:
					case KEY_A:
						sy+=4;
					break;

					case KEY_LEFT:
					case KEY_O:
						sx-=1;
					break;

					case KEY_RIGHT:
					case KEY_P:
						sx+=1;
					break;

					case KEY_Z:
						sx=0;
						sy=0;
						zm=1-zm;
					break;
				}

			}while(abort==0);
			// -------------	
		
		}else{
			TRACE("Aborting...");
		}
		
		// Safe to leave here as freeing NULL (if not enough space) should be okay...
		free(scr);
	}

	/**/
	TRACE("Finished.\n");
}
