//
// CVI Decoder 
//
// 02/09/2023: Uncompressed much faster (simple load/copy)
// 23/09/2023: Should fail more gracefully if not enough memory
// 23/09/2023: Removed stray 'free'
//

#include <cybiko.h>
#include "storm.h"

struct cvi_header{
	char id[4];
	int x;
	int y;
	char format[4];
	int type;
	char spare[4];
	int size;
}cvi_in;

/* */
int waitinput(struct module_t *);
void plot(char *scr,char *dis,long w,long h,long sx,long sy);
void zoom(char*,char*,int,int,int,int);
void gauge(struct module_t *,long,long);

// Function for huffman blocks
int codematch(long ,long *,short,int );
void gencode(char *);

extern long vb[];
extern char va[];

/* ************************ */

struct FileInput cvi_file_input;

void cvi(char *myfile,struct module_t *ptr_main_module){
	int bx,by,sx,sy,i,cnt,outcode;
	long pos,inbyt,*code,l;
	char *scr,*dis,*table,zm,abort;
	short tablesize;

	dis=DisplayGraphics_get_page_ptr(ptr_main_module->m_gfx, 0); 

	memset(dis,0,4000);

	// Use screen as temp buffer
	table=dis;		// 256 bytes
	code=dis+500;	// 1024 bytes

	/* ------------------------- */
	TRACE("Loading");

	FileInput_ctor(&cvi_file_input);

	if( !FileInput_open( &cvi_file_input,myfile  ) ){
		#if DEBUG==1
			TRACE("cvi: Could not open file.");
		#endif
	}else{
		Input_read( &cvi_file_input, &cvi_in,20);

		bx=cvi_in.x;
		by=cvi_in.y;

		scr=(char *)malloc( (bx>>2)*by);

		if (scr==0){
			TRACE("cvi: Could not allocate screen memory, needed %ld",(bx>>2)*by );
			TRACE("Size:%ld",cvi_in.size);
			TRACE("%ld x %ld (size:%ld)",bx,by,cvi_in.size);
		}else{
			
			switch(cvi_in.type){
			case 0:
				TRACE("Uncompressed (%ld)",cvi_in.type);
				//	
				FileInput_read( &cvi_file_input,scr,cvi_in.size);
			break;
			
			case 1:
				TRACE("Full RLE (%ld)",cvi_in.type);
				//
				pos=0;
				for(i=0;i<cvi_in.size;i+=2){
					char c1=FileInput_read_byte( &cvi_file_input);
					char c2=FileInput_read_byte( &cvi_file_input);
				
					memset( (scr+pos),c2,c1);
					pos+=c1;
				}
				//
			break;
			
			case 2:
				TRACE("Delta 1 (%ld)",cvi_in.type);
				memset( scr, cvi_in.spare[0], (bx>>2)*by);
				//
				pos=0;
				cnt=0;
				while(cnt<cvi_in.size){
					char c1=FileInput_read_byte( &cvi_file_input);
					cnt++;
				
					if (c1>=0){
						char c2=FileInput_read_byte( &cvi_file_input);
						for(i=0;i<c1;i++){
							scr[pos++]=c2;
						}
						cnt++;
					}else{
						pos-=c1;
					}	
				}
				//
			break;
			
			case 3:
				TRACE("Delta 2 (%ld)",cvi_in.type);
				memset( scr, cvi_in.spare[0], (bx>>2)*by);
				//
				pos=0;
				cnt=0;
				while(cnt<cvi_in.size){
					char c1=FileInput_read_byte( &cvi_file_input);
					cnt++;
					if (c1>=0){
						for(i=0;i<c1;i++){
							char c2=FileInput_read_byte( &cvi_file_input);
							scr[pos++]=c2;
						}
						cnt+=c1;
					}else{
						pos-=c1;
					}	
				}
				//
			break;

			case 4:
				//Read table to screen
				memset(dis,0,4000);
				inbyt=0;
				pos=0;
				tablesize=cvi_in.spare[3]+( (cvi_in.spare[2] & 0xff)<<8);
				FileInput_read(&cvi_file_input,table,tablesize);
				
				TRACE("Huffman (%ld) [%d]",cvi_in.type,tablesize);
			
				// Generate huff table to screen
				gencode(code);
				TRACE("Generated Table");
			
				//Extract to bitmap
				l=0;
				while(cnt<cvi_in.size){
					char c1=FileInput_read_byte( &cvi_file_input);
					for(i=0;i<8;i++){
						if ((c1 & va[i])==va[i] ){
							inbyt|=vb[l];
						}
			
						l++;
			
						if (l>1){			
							outcode=codematch(inbyt,code,tablesize,l);
							if (outcode>=0){
								inbyt=0;
								l=0;
								scr[pos++]=table[outcode];
							}
			
							if (l>33){
								TRACE("Code Length error!!");
							}
						}
					}
				
					gauge(ptr_main_module,cnt,cvi_in.size);		
					cnt++;
				}
			break;

			default:
				TRACE("Unknown (%ld)",cvi_in.type);
			break;
			}

			TRACE("Displaying. (%ld x %ld)",bx,by);

			abort=0;
			sx=0;
			sy=0;
			zm=0;

			// -----------
			do{
				if (sx<0){sx=0;}
				if (sy<0){sy=0;}

				if (zm==1){
					if (sx>((bx>>2)-20)){sx=(bx>>2)-20;}
					if ((sy+50)>by){sy=by-50;}

					// Show zoomed
					zoom(scr,dis,bx,by,sx,sy);
				}else{
					if ((sx<<2)+160>bx){sx=(bx>>2)-40;}
					if ((sy+100)>by){sy=by-100;}

					// Show normal
					plot(scr,dis,bx,by,sx,sy);
				}
				 
				DisplayGraphics_show( ptr_main_module->m_gfx );
					
				switch(waitinput(ptr_main_module)){
					case KEY_ESC:
						abort=1;
					break;

					case KEY_UP:
					case KEY_Q:
						sy-=4;
					break; 

					case KEY_DOWN:
					case KEY_A:
						sy+=4;
					break;

					case KEY_LEFT:
					case KEY_O:
						sx-=1;
					break;

					case KEY_RIGHT:
					case KEY_P:
						sx+=1;
					break;

					case KEY_Z:
						sx=0;
						sy=0;
						zm=1-zm;
					break;
				}
			}while(abort==0);
			
			
		}
		free(scr);
	// -----------    
	}

	FileInput_dtor( &cvi_file_input, LEAVE_MEMORY);

	TRACE("Finished.\n");

}
